rem -E disable natlink
rem needs elevated mode
C:\python25\python.exe D:\natlink\natlink\confignatlinkvocolaunimacro\natlinkconfigfunctions.py -E
C:\python26\python.exe D:\natlink\natlink\confignatlinkvocolaunimacro\natlinkconfigfunctions.py -E
