import D_

#Predefined for training purposes

Numbers = ["zero", "numeral", "0",'1','2','3','4','5','6','7','8','9','10','11','12','13',
      '14','15','16','17','18','19','20','25','30','35','40','45','50']
AppStarts = ["hyper-active",
    "Switch to", "Start TextPad", "Start BibPad", "Start Email", "Start TeX", "Start Studio New",
             "Start Netscape Communicator", "Start MATLAB", "Start APL", "Start M.S. Word",
             "Start Python Windows", "Start Bib Database", "Start FTP Pro",
             "Switch to Eudora"]
TeXFundamentals = ["Bib TeX mode", "stop Bib TeX mode", "TeX mode", "stop TeX mode",
                   "replace TeX signs",
                   'Bib Field','Next Field','Next Number','Next Pages','Bib Number',
                   'Get','get TeX command list','TeX','get TeX environment list','TeX around',
                   'Symbol','Term','Get Symbol list','Get Term list','Get Math Symbol list',
                   'Insert Equation','Inline Equation','Insert','Search','Get reference list',
                   'Search reference list','Search All In Title','Search And','Get All In Title',
                   'Get And','Search Some In Title','Search Or','Get Some In Title','Get Or'
                    ]                   
CommonCommands = ["Fix It", "Fix Discrete", "What Can I Say",
                  "Select That", "Spell That", "Cut That", "Delete That", "Scratch That",
                  "Cap That", "\\All-Caps That", "\\No-Caps That", "Compound That",
                  "Hyphenate That", "Format That", "Copy That",
                  "Resume With",
                  "Next", "Forward", "Previous", "Back", "Left", "Right", "Top", "Bottom",
                  "Beginning", "End",
                  "Word", "Words", "Character", "Characters", "Line", "Lines", "Paragraph","Paragraphs",
                  "Select Line", "Go to Top", "Go to Bottom",
                  "Insert Before That", "Insert After That", "Select Again",
                  "Choose 1", "Choose 2", "Choose 3",
                  "Choose 4", "Choose 5", "Choose 6",
                  "Choose 7", "Choose 8", "Choose 9", "Choose 10",
                  "Select 1", "Select 2", "Select 3",
                  "Select 4", "Select 5", "Select 6",
                  "Select 7", "Select 8", "Select 9", "Select 10",
                  "Switch to", "Start"]

SpecialTraining={
    'Numbers':          Numbers,
    'AppStarts':        AppStarts,
    'TeXFundamentals':  TeXFundamentals,
    'CommonCommands':   CommonCommands,
    'Alphabet':         D_.Alphabet,
    'ICAlphabet':       D_.ICAlphabet,        
    }